import sys
import os
from huggingface_hub import login, create_repo, upload_file
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from cusGPTdataset import Dataset
from cusGPTmodel import ModelGPT
try:
    from cusGPTvoice import Voice
except:
    print("WARNING: The voice function is not working as expected... Please wait until it's fixed")

class Trainer():
    def __init__(self,mix_pre=False,use_default=False,ds_name=None) -> None:
        self.dataset = Dataset(ds_name=ds_name)
        self.model = ModelGPT(mix_pre,use_default)
        self.login_hf_code = None

    def login_hf(self,token):
        self.login_hf_code = token
        login(token)

    def data(self,file_name,first_n=None,start_from=0,name="WICKED4950/MentalHeathEsther",local=False,answer_only=True,batch_size=8,maskq=False,toe_len=512):
        if self.model.tokenizer == None:
            raise Exception("You have to load the model first because this functions uses the tokneizer of that model")
        else:
            self.dataset(file_name=file_name,tokenizer = self.model.tokenizer,first_n=first_n,start_from=start_from,name=name,local=local,answer_only=answer_only,batch_size=batch_size,maskq=maskq,toe_len=toe_len)

    def initVoice(self,voice_name=""):
        global voice
        voice = Voice(voice_name)

    def train(self,epochs=1,val_dataset=None, lr=5e-5, final_lr=5e-5,loss_avg=1500,change_batch=500,split_training=False,val_split=0,val_log=500,batch_acc=1,use_v2=False):
        if self.dataset.return_tf() == None:
            raise ValueError("You have to first load and process the data to train the model!")
        else:
            if use_v2:
                self.model.train_modelV2(self.dataset.return_tf(),val_dataset, epochs,lr, final_lr, loss_avg, change_batch=change_batch,split_training=split_training, val_split=val_split,val_log=val_log,batch_acc=batch_acc)
            else:
                self.model.train_model(self.dataset.return_tf(),val_dataset, epochs,lr, final_lr, loss_avg, change_batch=change_batch,split_training=split_training, val_split=val_split,val_log=val_log,batch_acc=batch_acc)
            print("training done!")

    def push_model(self,name="saved_model",resp_name=None):
        if self.login_hf_code == None:
            raise ValueError("You have to log in into your hf account using .login_tf('Code here')")
        if resp_name == None:
            raise Exception("You have to give a name for the place to save the model")
        self.model.save_model(name)
        self.model.push_model(resp_name)


def dataset_obj(ds_name):
    return Dataset(ds_name=ds_name)


from tqdm import tqdm
import tensorflow as tf
import shutil

class Dataset():
    def __init__(self,ds_name=None):
        self.questions = None
        self.answers = None
        self.gpt_sentence = None
        self.tf_dataset = None
        self.type = 0
        if ds_name == None:
            self.dt_save_path = "encoded_dataset"
        else:
            self.dt_save_path = ds_name


    def dataset_load(self,file_name,first_n=None,start_from=0,name="WICKED4950/MentalHeathEsther",local=True):
      if local:
          import json
          with open(f"{file_name}.json") as f:
              data = json.load(f)
          if first_n == None:
              first_n = len(list(data.keys()))
          self.questions = list(data.keys())[start_from:first_n]
          self.answers = list(data.values())[start_from:first_n]
      else:
          try:
              from datasets import load_dataset
          except:
              raise RuntimeError("You didn't run the pip install at the start of the program")
          dataset = load_dataset(name, data_files=file_name)
          train_data_dict = dataset['train'].to_dict()
          if first_n == None:
              first_n = len(train_data_dict["question"])
          self.questions = train_data_dict["question"][start_from:first_n]
          self.answers = train_data_dict["answer"][start_from:first_n]
          self.type = 1

    def datasetprocess(self,user_start="<|SOH|>",bot_start="<|SOB|>",answer_only=False):
        if answer_only:
            if self.questions == None:
                raise ValueError("Please do the dataset_loadqa() to get the dataset running and then process it")
            if user_start == "<|SOH|>" and bot_start=="<|SOB|>":
                self.gpt_sentence = self.answers
            else:
                correct_lab_list = []
                for sentence in self.answer:
                    sentence = sentence.replace("<|SOH|>", user_start)
                    sentence = sentence.replace("<|SOB|>", bot_start)
                self.gpt_sentence = correct_lab_list
        else:
            if self.questions == None:
                raise ValueError("Please do the dataset_loadqa() to get the dataset running and then process it")

            sentence = [f"{user_start}{question}{bot_start}{answer}{user_start}" for question,answer in zip(self.questions,self.answers)]
            self.gpt_sentence = sentence
        self.type = 2

    def mask_qa(self,tokenizer ,conversations ,len_=1024,batch_size=8):
        all_input_ids = []
        all_attention_masks = []
        all_labels = []
        tokenized_output = tokenizer(
                conversations,
                return_tensors='tf',
                padding='max_length',  # Pad to the maximum length
                truncation=True,       # Truncate if longer than max_length
                max_length=len_  # Set the maximum sequence length
            )
        
        soh = tokenizer.encode("<|SOH|>", add_special_tokens=False)[0]
        sob = tokenizer.encode("<|SOB|>", add_special_tokens=False)[0]
        for index in tqdm(range(tokenized_output["input_ids"].shape[0]), desc="Processing conversations", unit="conversation"):

            input_ids = tokenized_output["input_ids"][index]
            attention_mask = tokenized_output["attention_mask"][index]
            # Create the labels
            labels = input_ids.numpy().copy()  # Make sure it's a NumPy array for easier manipulation


            labels[0] = -100  # Start by masking the first token
            inp_cpy = input_ids.numpy()
            last = len(input_ids.numpy()) - 2
            for idx, token in enumerate(input_ids.numpy()):
                if attention_mask[idx] == 0:
                    break
                if token == soh:  # If it's a question start token, mask the subsequent tokens
                    try:
                        while inp_cpy[idx] != sob:
                            labels[idx + 1] = -100
                            idx += 1
                            if idx == last:
                                break
                    except:
                        pass


            # Append input_ids, attention_mask, and labels to their respective lists
            all_input_ids.append(input_ids)
            all_attention_masks.append(attention_mask)
            all_labels.append(labels)

        dataset = self.create_tf_dataset(tf.concat(tf.convert_to_tensor(all_input_ids), axis=0),tf.concat(tf.convert_to_tensor(all_attention_masks), axis=0),tf.concat(tf.convert_to_tensor(all_labels), axis=0))
        dataset = dataset.shuffle(buffer_size=75000)  # Shuffle with buffer size of 1000
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
        self.tf_dataset = dataset
        self.type = 3

    def create_tf_dataset(self,inputs,attention_mask=None,labels=None):
        if labels == None:
            if attention_mask == None:
                input_ids = inputs['input_ids']
            else:
                input_ids = inputs
            if attention_mask == None:
                attention_mask = inputs['attention_mask']
            return tf.data.Dataset.from_tensor_slices(({
                'input_ids': input_ids,
                'attention_mask': attention_mask
            }, input_ids))
        else:
            return tf.data.Dataset.from_tensor_slices(({
                'input_ids': inputs,
                'attention_mask': attention_mask
            }, labels))

    def tokenize(self, tokenizer, max_length=256, batch_size=8):
        if self.gpt_sentence is None:
            raise ValueError("Please run dataset_loadqa() and datasetgpt() first before processing.")

        # Tokenize the GPT sentences
        inputs = tokenizer(self.gpt_sentence, padding=True, truncation=True, return_tensors='tf', max_length=max_length)


        # Create the dataset and apply shuffle and batching
        dataset = self.create_tf_dataset(inputs)
        dataset = dataset.shuffle(buffer_size=75000)  # Shuffle with buffer size of 1000
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
        self.tf_dataset = dataset
        self.type = 3

    def clear_save(self):
        try:
            shutil.rmtree(self.dt_save_path)
            self.tf_dataset = None
        except FileNotFoundError:
            print("Brother! There is no dataset to delete... But I forgive you and not going to throw a error")
        except Exception as e:
            print(f"An unexpected error occurred: {e}")

    def __call__(self,file_name,tokenizer,first_n=None,start_from=0,name="WICKED4950/MentalHeathEsther",local=False,answer_only=True,batch_size=8,maskq=False,toe_len=512):
        self.tokenizer = tokenizer
        try:
            self.tf_dataset = tf.data.Dataset.load(self.dt_save_path)
            self.return_tf()
        except:
            self.dataset_load(file_name=file_name,first_n=first_n,start_from=start_from,name=name,local=local)
            self.datasetprocess(answer_only=answer_only)
            if maskq:
                self.datasetprocess(answer_only=answer_only)
                self.mask_qa(conversations=self.gpt_sentence,tokenizer=tokenizer,len_=toe_len,batch_size=batch_size)
            else:
                self.tokenize(tokenizer=tokenizer,max_length=toe_len,batch_size=batch_size)
            tf.data.Dataset.save(self.tf_dataset, self.dt_save_path)


    def __len__(self):
        try:
            return len(self.gpt_sentence)
        except:
            return 0

    @property
    def __class__(self):
        return Dataset

    def return_tf(self):
        return self.tf_dataset


from setuptools import setup, find_packages

setup(
    name="customGPTBW",
    version="0.1.49",
    packages=find_packages(),
    install_requires=[
        "tensorflow",
        "transformers",
        "tqdm",
        "psutil",
        "huggingface_hub",
        "datasets",
        "numpy"
    ],
    description="A private module for dataset loading and GPT training.",
    author="Boring._.wicked",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
    ],
)


from tqdm import tqdm
import tensorflow as tf
import threading
import multiprocessing as mp
from functools import partial
import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor
import numpy as np
import time

class Timer():
    def __init__(self) -> None:
        self.starting_time = None

    def timer(self,func,*args,**kwargs):
        starting_time = time.time()
        result = func(*args,**kwargs)
        print(f"Time taken to execute {func.__name__} is {time.time() - starting_time:.4f}s")

    def start(self):
        self.starting_time = time.time()

    def end(self):
        print(f"Time taken to execute is {time.time() - self.starting_time:.4f}s")

def tokenize_batch(batch, tokenizer, max_length):
    return tokenizer(batch, padding='max_length', truncation=True, return_tensors='tf', max_length=max_length)

def split_into_chunks(lst, n_chunks):
    n = len(lst)
    n_chunks = min(n, n_chunks)
    k, m = divmod(n, n_chunks)
    return [lst[i * k + min(i, m): (i + 1) * k + min(i + 1, m)] for i in range(n_chunks)]

def tokenize(ds_obj, mod_obj, max_length=512, batch_size=8, num_cores=96):
    if ds_obj.gpt_sentence is None:
        raise ValueError("Please run dataset_loadqa() and datasetgpt() first before processing.")

    sentences = ds_obj.gpt_sentence

    # Split into parallel batches
    batches = split_into_chunks(sentences, num_cores)

    # Prepare a partial function that already has the tokenizer and max_length parameters set.
    tokenize_func = partial(tokenize_batch, tokenizer=mod_obj.tokenizer, max_length=max_length)

    # Process batches in parallel
    with mp.Pool(num_cores) as pool:
        batch_results = pool.map(tokenize_func, batches)

    # Aggregate results
    all_inputs, all_masks = [], []
    for result in batch_results:
        all_inputs.extend(result['input_ids'])
        all_masks.extend(result['attention_mask'])

    input = {"input_ids":tf.convert_to_tensor(all_inputs),"attention_mask":tf.convert_to_tensor(all_masks)}
    dataset = ds_obj.create_tf_dataset(input)
    dataset = dataset.batch(batch_size).prefetch(tf.data.AUTOTUNE)
    ds_obj.tf_dataset = dataset
    
def multi_run(tasks:list):
    threads = []
    for func in tasks:
        t = threading.Thread(target=func)
        t.start()
        threads.append(t)

    # Wait for all threads to finish
    for t in threads:
        t.join()



import tensorflow as tf
from transformers import TFAutoModelForCausalLM, AutoTokenizer, AutoConfig
import sys
import psutil
from tqdm import tqdm


class ModelGPT():
    """
    The class that handles and holds the model and model related custom functions
    """

    def error_handler(func):
        """
        Just used for the training function so that it catches the last error when training the model and the last graph hits... Usefulll only when using the splited training.
        """
        def wrapped(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except:
                pass
        return wrapped
    

    def __init__(self,mix_pre=False,use_default=False):
        """
        Create a instance of the ModelGPT class for handling functions and loaded model
        
        Args-
        - mix_pre -> Set this to true if you want to use 'mixed_bfloat16'
        - use_default -> Makes it so that the code uses the avaliable cpu rather than a tpu or gpu
        """
        self.model = None
        self.tokenizer = None
        self.saved_model_name = None
        self.context = []
        self.all_train_loss = []
        tf.config.optimizer.set_jit(True)
        try:
            if use_default:
                try:
                    tf.config.threading.set_intra_op_parallelism_threads(24)
                    tf.config.threading.set_inter_op_parallelism_threads(2)
                except:
                    pass
                raise ValueError("skip")
            resolver = tf.distribute.cluster_resolver.TPUClusterResolver()
            tf.config.experimental_connect_to_cluster(resolver)
            tf.tpu.experimental.initialize_tpu_system(resolver)
            self.strategy = tf.distribute.TPUStrategy(resolver)
            print("TPU connected")
        except ValueError:
            self.strategy = tf.distribute.MultiWorkerMirroredStrategy()
            print("Using default strategy")
        if mix_pre:
            policy = tf.keras.mixed_precision.Policy('mixed_bfloat16')
            tf.keras.mixed_precision.set_global_policy(policy)

    def load_gptmodel(self,name="gpt2",grad_check=False):
        """
        Used to load a gpt model from huggining face. You have to logged in the use private models and tokenizers
        
        Args-
        - name -> Give in the name of the resp from hugging face to load the model and the tokenizer
        - grad_check -> Set it to true if you want to load the gradentient check point comaptable model
        """
        from huggingface_hub import login, create_repo, upload_file

        self.tokenizer = AutoTokenizer.from_pretrained(name)
        self.tokenizer.pad_token = self.tokenizer.eos_token
        with self.strategy.scope():
            if grad_check:
                config = AutoConfig.from_pretrained(name)
                config.gradient_checkpointing = True
                self.model = TFAutoModelForCausalLM.from_pretrained(name, config=config)
            else:
                self.model = TFAutoModelForCausalLM.from_pretrained(name)
    
    def load_custom_model(self,func,*args,**kwargs):
        """
        Used to load a custom model build and ready to be used (Must be a object compatable with every functions)
        
        Args-
        - model -> The model object on which the code can work on.
        """
        with self.strategy.scope():
            self.model = func(*args,**kwargs)
    
    def load_custom_tokenizer(self,name):
        """
        For now this function is used to load a tokenizer from hugging face alone without loading the model.
        
        Args-
        - name -> Name of the resp from which the function should load the tokenizer only.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(name)
        self.tokenizer.pad_token = self.tokenizer.eos_token

    def add_tokens(self,new_token_list):
        """
        Used to add new tokens into the model and the tokenizer. Both the tokeizer and the model should be loaded
        
        Args-
        - new_token_list -> Takes in a input of the type list containing string element to add the the tokenizer and model
        """
        self.tokenizer.add_tokens(new_token_list)
        with self.strategy.scope():
            self.model.resize_token_embeddings(len(self.tokenizer))

    def freeze(self,value,shift=0,mix=False):
        """
        Used to freeze and unfreeze blocks of models in patterns
        
        Args-
        - value -> Takes in a int to specifiy in which multiple the models layer will be unfreezed in
        - shift -> Used to add a integer value to add the the multiple to shift the unfreezen layer
        - mix -> Set this to True if you want to make it so that the function freeze blocks in patterns or the blocks will be unfreezed from the last
        """
        if mix:
            for i, layer in enumerate(self.model.transformer.h):#self.model.get_layer(index=0).decoder.layers
                layer.trainable = False
                if ((i+shift)%value) ==0:
                    # Set the entire layer to not trainable
                    layer.trainable = True
        else:
            for i, layer in enumerate(self.model.transformer.h):
                layer.trainable = False
            for i, layer in enumerate(reversed(self.model.transformer.h)):
                if i == value:
                    break
                layer.trainable = True

        # Verify the frozen status of the layers
        """for i, layer in enumerate(self.model.transformer.h):
            print(f"Layer {i} {'Frozen' if not layer.trainable else 'Trainable'}")"""


    def train_model(self, dataset_,val_dataset=None ,epochs=1, lr=5e-5, final_lr=5e-5, loss_avg=1500, change_batch=500, split_training=False, val_split=0.1,val_log=500,batch_acc=1):
        """Used to train the loaded model
        
        Args-
        - dataset_ -> Main dataset to be used for the training (should be a tf.dataset)
        - val_dataset -> The validation dataset to be used
        - loss_avg -> Takes the loss of n number of the past loss from batch and calculates their average to be displayed
        - (NOT CREATED ANYMORE)"""

        val = True
        strategy = self.strategy

        @tf.function
        def train_step(batch):
            input_ids = batch[0]['input_ids']
            attention_mask = batch[0]['attention_mask']
            labels = batch[1]

            with tf.GradientTape() as tape:
                outputs = self.model(input_ids, attention_mask=attention_mask, labels=labels)
                loss = outputs.loss

            gradients = tape.gradient(loss, self.model.trainable_variables)
            clipped_gradients = [tf.clip_by_value(grad, -1.0, 1.0) for grad in gradients]
            optimizer.apply_gradients(zip(clipped_gradients, self.model.trainable_variables))
            return loss


        @tf.function
        def val_step(batch):
            input_ids = batch[0]['input_ids']
            attention_mask = batch[0]['attention_mask']
            labels = batch[1]

            outputs = self.model(input_ids, attention_mask=attention_mask, labels=labels)
            loss = outputs.loss
            return loss 

        dataset_size = len(dataset_)
        self.dataset_size = dataset_size
        if val_split == 0:
            val = False
        if val_dataset != None:
            train_dataset = dataset_
        else:
            val_size = int(dataset_size * val_split)
            train_size = dataset_size - val_size
            train_dataset = dataset_.take(train_size)
            val_dataset = dataset_.skip(train_size)
        if val_log > dataset_size:
            val_log = dataset_size

        initial_lr = lr
        decay_rate = (final_lr / initial_lr) ** (1 / (dataset_size*epochs))
        lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
            initial_learning_rate=lr,
            decay_steps=1,
            decay_rate=decay_rate,
            staircase=False  # Smooth decay (not step-wise)
        )

        def val_func(strategy,dist_val_dataset):
            all_val_loss = []
            for batch in dist_val_dataset:
                per_replica_losses = strategy.run(val_step, args=(batch,))
                loss = strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None)

                if tf.math.is_nan(loss):
                    continue

                all_val_loss.append(loss.numpy())
                val_losses = all_val_loss

                # Print the loss and overwrite it on each iteration
                avg_val_loss = float(sum(val_losses) / len(val_losses))
                sys.stdout.write(f'\r Calculating.... Validation Loss: {avg_val_loss:.4f}')  # Overwrite the previous print
                sys.stdout.flush()

            print("")
                

            
        with strategy.scope():
            # Distribute datasets
            dist_train_dataset = strategy.experimental_distribute_dataset(train_dataset)
            if val:
                dist_val_dataset = strategy.experimental_distribute_dataset(val_dataset)

            optimizer = tf.keras.optimizers.AdamW(learning_rate=lr_schedule)
            all_batch_count = 0
            for epoch in range(epochs):
                all_val_loss = []

                # Training Loop
                progress_bar = tqdm(dist_train_dataset, desc=f"Epoch {epoch + 1} [Training]", unit="batch")
                i = 0
                j = 0
                for batch in progress_bar:
                    if split_training:
                        if (j % change_batch) == 0:
                            print("switching")
                            self.freeze(4, shift=i, mix=True)
                            i += 1
                        j += 1
                    
                    all_batch_count = all_batch_count + 1
                    if ((all_batch_count%val_log) == 0) and (val_log != dataset_size) and (val == True):
                        val_func(strategy=strategy,dist_val_dataset=dist_val_dataset)
                    per_replica_losses = strategy.run(train_step, args=(batch,))
                    loss = strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None)

                    if tf.math.is_nan(loss):
                        continue

                    self.all_train_loss.append(loss.numpy())
                    train_losses = self.all_train_loss[-loss_avg:]

                    memory_used = psutil.virtual_memory().used / (1024 ** 3)  # in GB
                    cpu_percent = psutil.cpu_percent(interval=0)  # in percentage

                    progress_bar.set_postfix(loss=(sum(train_losses) / len(train_losses)), memory=f"{memory_used:.2f} GB", cpu=f"{cpu_percent}%", inst_loss=(train_losses[-1]),perplexity=(tf.exp(sum(train_losses) / len(train_losses)).numpy()))
                
                if (val_log == dataset_size) and (val == True):
                    val_func(strategy=strategy,dist_val_dataset=dist_val_dataset)

        return self.all_train_loss


    def train_modelV2(self, dataset_,val_dataset=None ,epochs=1, lr=5e-5, final_lr=5e-5, loss_avg=1500, change_batch=500, split_training=False, val_split=0.1,val_log=500,batch_acc=1):
        val = True
        strategy = self.strategy

        @tf.function
        def train_step(batch):
            input_ids = batch[0]['input_ids']
            attention_mask = batch[0]['attention_mask']
            labels = batch[1]

            with tf.GradientTape() as tape:
                loss = self.model(input_ids, attention_mask=attention_mask, labels=labels).loss
            return loss , tape.gradient(loss, self.model.trainable_variables)
        
        @tf.function
        def apply_grad(list_gradients):
            avg_grads = [tf.math.add_n(grads) for grads in zip(*list_gradients)]
            avg_grads = [tf.math.divide(g, tf.cast(batch_acc, tf.float32)) for g in avg_grads]
            optimizer.apply_gradients(zip(avg_grads, self.model.trainable_variables))

        @tf.function
        def val_step(batch):
            return self.model(
                batch[0]['input_ids'], 
                attention_mask=batch[0]['attention_mask'], 
                labels=batch[1]
            ).loss

        dataset_size = len(dataset_)
        self.dataset_size = dataset_size
        if val_split == 0:
            val = False
        if val_dataset != None:
            train_dataset = dataset_
        else:
            val_size = int(dataset_size * val_split)
            train_size = dataset_size - val_size
            train_dataset = dataset_.take(train_size)
            val_dataset = dataset_.skip(train_size)
        if val_log > dataset_size:
            val_log = dataset_size

        initial_lr = lr
        decay_rate = (final_lr / initial_lr) ** (1 / (dataset_size*epochs))
        lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
            initial_learning_rate=lr,
            decay_steps=1,
            decay_rate=decay_rate,
            staircase=False  # Smooth decay (not step-wise)
        )

        def val_func(strategy,dist_val_dataset):
            all_val_loss = []
            for batch in dist_val_dataset:
                per_replica_losses = strategy.run(val_step, args=(batch,))
                loss = strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None)

                if tf.math.is_nan(loss):
                    continue

                all_val_loss.append(loss.numpy())
                val_losses = all_val_loss

                avg_val_loss = float(sum(val_losses) / len(val_losses))
                sys.stdout.write(f'\r Calculating.... Validation Loss: {avg_val_loss:.4f}')
                sys.stdout.flush()

            print("")
                

            
        with strategy.scope():
            optimizer = tf.keras.optimizers.AdamW(learning_rate=lr_schedule)
            dist_train_dataset = strategy.experimental_distribute_dataset(train_dataset)
            if val:
                dist_val_dataset = strategy.experimental_distribute_dataset(val_dataset)

            all_batch_count = 0
            for epoch in range(epochs):
                all_val_loss = []

                # Training Loop
                progress_bar = tqdm(dist_train_dataset, desc=f"Epoch {epoch + 1} [Training]", unit="batch")
                i = 0
                j = 0
                list_gradients = []
                for batch in progress_bar:
                    if split_training:
                        if (j % change_batch) == 0:
                            print("switching")
                            self.freeze(4, shift=i, mix=True)
                            i += 1
                        j += 1
                    
                    all_batch_count = all_batch_count + 1
                    if ((all_batch_count%val_log) == 0) and (val_log != dataset_size) and (val == True):
                        val_func(strategy=strategy,dist_val_dataset=dist_val_dataset)
                    per_replica_losses, per_replica_gradients = strategy.run(train_step, args=(batch,))
                    loss = strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None)
                    list_gradients.append([strategy.reduce(tf.distribute.ReduceOp.MEAN, g, axis=None) for g in per_replica_gradients])

                    if all_batch_count%batch_acc == 0:
                        strategy.run(apply_grad,args=(list_gradients,))
                        list_gradients = []
                    
                    if tf.math.is_nan(loss):
                        continue

                    self.all_train_loss.append(loss.numpy())
                    train_losses = self.all_train_loss[-loss_avg:]

                    memory_used = psutil.virtual_memory().used / (1024 ** 3)  # in GB
                    cpu_percent = psutil.cpu_percent(interval=0)  # in percentage

                    progress_bar.set_postfix(loss=(sum(train_losses) / len(train_losses)), memory=f"{memory_used:.2f} GB", cpu=f"{cpu_percent}%", inst_loss=(train_losses[-1]),perplexity=(tf.exp(sum(train_losses) / len(train_losses)).numpy()))
                
                if (val_log == dataset_size) and (val == True):
                    val_func(strategy=strategy,dist_val_dataset=dist_val_dataset)

        return self.all_train_loss

    def save_model(self,name="path_to_save_your_model"):
        self.saved_model_name = name
        self.model.save_pretrained(name)
        self.tokenizer.save_pretrained(name)

    def upload(self,rep_name,path = None):
        if path == None:
            path = self.saved_model_name
        from huggingface_hub import login, create_repo, upload_file
        import os
        files = os.listdir(path)
        model_files = []
        for file_ in files:
            model_files.append(f"/content/{path}/{file_}")
        repo_path = rep_name

        for model_file in model_files:
            upload_file(
                path_or_fileobj=model_file,
                path_in_repo=model_file.split("/")[-1],
                repo_id=repo_path,
                repo_type="model",
            )
    def push_model(self,name):
        from huggingface_hub import login, create_repo, upload_file
        repo_url = create_repo(name)
        print(f"Repository created at: {repo_url}")
        self.upload(name)

    @property
    def get_context(self):
        return "".join(f"<|SOH|>{msg[0]}<|SOB|>{msg[1]}" for msg in self.context)

    def generate_text(self, prompt, sos="<|SOH|>", eos="<|SOB|>", recall_sent=4,sys_prom=None,show_rawinp=False,ret_sq=1,temperature=0.5,top_k=2):
        if prompt == "!clear":
            self.context = []
            return "Context cleared."

        input_text = "".join(f"{sos}{msg[0]}{eos}{msg[1]}" for msg in self.context[-recall_sent:])
        if sys_prom == None:
            pass
        else:
            input_text = "<|SYS|>"+sys_prom+input_text
        input_text += f"{sos}{prompt}{eos}"
        if show_rawinp:
            print(input_text)

        raw_responses = self.raw_pred(input_text, max_length=512,ret_sq=ret_sq,temperature=temperature,top_k=top_k)
        raw_responses = [raw_response.replace(" <|SOH|> ","<|SOH|>") for raw_response in raw_responses]
        raw_responses = [raw_response.replace("<|SOH|> ","<|SOH|>") for raw_response in raw_responses]
        raw_responses = [raw_response.replace(" <|SOH|>","<|SOH|>") for raw_response in raw_responses]
        raw_responses = [raw_response.replace(" <|SOB|> ","<|SOB|>") for raw_response in raw_responses]
        raw_responses = [raw_response.replace(" <|SOB|> ","<|SOB|>") for raw_response in raw_responses]
        raw_responses = [raw_response.replace(" <|SOB|> ","<|SOB|>") for raw_response in raw_responses]
        responses = [raw_response.split("<|SOB|>")[-1] for raw_response in raw_responses]
        responses = [raw_response.replace("<|SOH|>","") for raw_response in responses]
        if ret_sq != 1:
            count = 0
            for i in responses:
                print(count,i)
                count = count + 1
            index = input(": ")
            if index == "":
                index = 0
            try:
                self.context.append((prompt, responses[int(index)]))
            except:
                self.context.append((prompt, index))
        else:
            self.context.append((prompt, responses[0]))

        return responses

    def raw_pred(self, input, max_length=20, temperature=0.5,ret_sq=1,top_k=3):
        input_ids = self.tokenizer.encode(input, return_tensors='tf')

        # Create attention mask
        attention_mask = tf.ones_like(input_ids)

        with self.strategy.scope():
            outputs = self.model.generate(
                input_ids,
                attention_mask=attention_mask,  # Explicit attention mask
                max_length=input_ids.shape[1] + max_length,
                top_k=top_k,
                do_sample=True,
                temperature=temperature,
                pad_token_id=self.tokenizer.eos_token_id,
                eos_token_id=self.tokenizer.encode("<|SOH|>", add_special_tokens=False)[0],  # Custom EOS token ID
                num_return_sequences=ret_sq  # Specify how many sequences to generate
            )

        # Decode generated output
        generated_text = [self.tokenizer.decode(output, skip_special_tokens=True) for output in outputs]
        return generated_text